# Copyright (c) 2024, Thomas Goirand <zigo@debian.org>
# Copyright (c) 2024, Philippe Serafin <philippe.serafin@infomaniak.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This script check a mounted device is full, and disable
# its matching rsync module if that is the case. The disk
# full limit is the first argument.

import configparser
import io
import os
import shutil
import sys

from oslo_config import cfg
from oslo_log import log as logging

LOG = logging.getLogger(__name__, project='swift-drive-full-checker')

GiB = 1024 * 1024 * 1024

# Initialize script common oslo_config
common_opts = [
    cfg.IntOpt('account_max_connections', short='b', default=8,
               help='Max connections to the Account rsync backend.'),

    cfg.IntOpt('account_reserved_space', short='a', default=100,
               help='Account server reserved space in GiB.'),

    cfg.StrOpt('account_rsyncd_section_name', default=' account_{} ',
               help='Account section name in the rsyncd.conf file. '
               'The "{}" sign will be replaced by the drive name. '
               'If not using per drive sections, simply write "account". '
               'Note that spaces can be used, depending in your setup.'),


    cfg.IntOpt('container_reserved_space', short='c', default=100,
               help='Container server reserved space in GiB.'),

    cfg.IntOpt('container_max_connections', short='e', default=8,
               help='Max connections to the Container rsync backend.'),

    cfg.StrOpt('container_rsyncd_section_name', default=' container_{} ',
               help='Container section name in the rsyncd.conf file. '
               'The "{}" sign will be replaced by the drive name. '
               'If not using per drive sections, simply write "container". '
               'Note that spaces can be used, depending in your setup.'),

    cfg.IntOpt('object_reserved_space', short='o', default=100,
               help='Object server reserved space in GiB.'),

    cfg.IntOpt('object_max_connections', short='p', default=8,
               help='Max connections to the Object rsync backend.'),

    cfg.StrOpt('object_rsyncd_section_name', default=' object_{} ',
               help='Object section name in the rsyncd.conf file. '
               'The "{}" sign will be replaced by the drive name. '
               'If not using per drive sections, simply write "object". '
               'Note that spaces can be used, depending in your setup.'),

    cfg.StrOpt('storage_path', default='/srv/node',
               help='Mount point of your storage.'),

    cfg.StrOpt('rsyncd_conf_path', short='s', default='/etc/rsyncd.conf',
               help='Mount point of your storage.'),
]


def list_opts():
    return [
        ("DEFAULT", common_opts),
    ]


def configure_rsyncd_conf(account_rs, account_mc, account_secname,
                          container_rs, container_mc, container_secname,
                          object_rs, object_mc, object_secname,
                          storage_p, rsyncd_p):
    # Load the rsyncd.conf file, adding
    # a fake global section.
    fake_section = '[fake_section_to_please_configobj]\n'
    try:
        with open(rsyncd_p, 'r') as f:
            file_content = fake_section + f.read()
    except Exception as err:
        print("Unexpected error reading {}: {}".format(rsyncd_p, err))
        return 1

    cp = configparser.RawConfigParser()
    if sys.version_info[0] == 2:
        cp.read_string(file_content.decode('unicode-escape'))
    else:
        cp.read_string(file_content)

    # For all dirs in /srv/node
    for dir in os.listdir(storage_p):
        dirpath = os.path.join(storage_p, dir)
        # If the dir is mounted
        if os.path.ismount(dirpath):
            # shutil.disk_usage can be mocked in tests.
            if sys.version_info[0] == 2:
                space = os.statvfs(dirpath)
                free = (space.f_bsize * space.f_bavail)
            else:
                free = shutil.disk_usage(dirpath).free

            # Calculate section name (ie: replace '{}' by drive name)
            if '{}' in account_secname:
                search_str = account_secname.format(dir)
            else:
                search_str = account_secname

            # If referenced in the rsyncd.conf
            # In old setup (Python 2), calling config_parser['something']
            # raises an exception if the something section is not present
            # in the config file. Which is why we must do try/except.
            # I believe this try/except can be removed on more recent
            # Python3 based setups.
            try:
                if cp[search_str]:
                    # If partition is full, set 'max connections'
                    # to -1 to disable rsync
                    if free < account_rs:
                        cm = -1
                    else:
                        cm = account_mc

                    if int(cp[search_str]['max connections']) != cm:
                        if cm == -1:
                            LOG.warning('Disabling ' + search_str)
                        else:
                            LOG.warning('Enabling ' + search_str)
                        cp[search_str]['max connections'] = str(cm)
            except KeyError:
                pass

            # Same for container
            if '{}' in container_secname:
                search_str = container_secname.format(dir)
            else:
                search_str = container_secname

            try:
                if cp[search_str]:
                    if free < container_rs:
                        cm = "-1"
                    else:
                        cm = container_mc

                    if int(cp[search_str]['max connections']) != cm:
                        if cm == "-1":
                            LOG.warning('Disabling ' + search_str)
                        else:
                            LOG.warning('Enabling ' + search_str)
                        cp[search_str]['max connections'] = str(cm)
            except KeyError:
                pass

            # Same for object
            if '{}' in object_secname:
                search_str = object_secname.format(dir)
            else:
                search_str = object_secname

            try:
                if cp[search_str]:
                    if free < object_rs:
                        cm = -1
                    else:
                        cm = object_mc

                    if int(cp[search_str]['max connections']) != cm:
                        if cm == "-1":
                            LOG.warning('Disabling ' + search_str)
                        else:
                            LOG.warning('Enabling ' + search_str)

                        cp[search_str]['max connections'] = str(cm)
            except KeyError:
                pass

    # Prepare our rsyncd.conf file before writing
    iow = io.StringIO()
    cp.write(iow)
    file_out = iow.getvalue().replace(fake_section, '')

    try:
        with open(rsyncd_p, 'w') as f:
            f.write(file_out)
    except Exception as err:
        LOG.error("Unexpected error {}".format(err))
        return 1


def main():
    """
    Entry point for the script
    """

    CONF = cfg.ConfigOpts()
    CONF.register_cli_opts(common_opts)
    logging.register_options(CONF)

    try:
        CONF(sys.argv[1:], project='swift',
             default_config_files=['/etc/swift/drive-full-checker.conf'])
    except Exception as E:
        print(E)
        CONF.print_usage()
        return 1

    logging.setup(CONF, CONF.project)

    # Convert input to GiB
    account_reserved_space = CONF.account_reserved_space * GiB
    container_reserved_space = CONF.container_reserved_space * GiB
    object_reserved_space = CONF.object_reserved_space * GiB

    return configure_rsyncd_conf(account_reserved_space,
                                 CONF.container_max_connections,
                                 CONF.account_rsyncd_section_name,
                                 container_reserved_space,
                                 CONF.container_max_connections,
                                 CONF.container_rsyncd_section_name,
                                 object_reserved_space,
                                 CONF.object_max_connections,
                                 CONF.object_rsyncd_section_name,
                                 CONF.storage_path,
                                 CONF.rsyncd_conf_path)


if __name__ == "__main__":
    sys.exit(main())
